class TMap {
    private variable CName      ;# name of the map
    private variable CLevel     ;# level of the map
    private variable CId        ;# a unique id
    private variable CMapM      ;# the Map manager, that needs to be notified of changes
    private variable CParentWid ;# widget ID of the parent
    private variable CCanvasWid ;# widget ID of the scrolled canvas
    private variable CCBWid     ;# widget ID of the canvas

    private variable CMud

    private variable CRoomList  ;# this is an array of objects
    private variable CChanged

    private variable CXmin "-50"
    private variable CXmax "50"
    private variable CYmin "-50"
    private variable CYmax "50"


    variable CZoomX
    variable CZoomY
    variable CPlayerX
    variable CPlayerY
    variable CHasZoom
    variable CHasPlayer

    method xmin {} { return $CXmin }
    method xmax {} { return $CXmax }
    method ymin {} { return $CYmin }
    method ymax {} { return $CYmax }
    method size {} { return [list $CXmin $CXmax $CYmin $CYmax]}

    method resize {a b c d}

    private variable CUnLoaded 0

    private variable CBGIM
    private common CBGColor "\#daceaf"
    private common CGrid  32

    # 1: mapname, 2: maplevel
    constructor {args} {}
    destructor {}

    method speedWalkTo {a b} {
	if {[$CMapM roomHasChanged]} {
	    if {[$CMapM devMode]} {
		$CMapM acceptRoom
	    } else {
		$CMapM informRoomHasChanged
		return
	    }
	}
	
	set x [p2mi [$CCBWid canvasx $a]  1]
	set y [p2mi [$CCBWid canvasy $b] -1]
	$CMapM speedWalkTo $this $x $y
    }

    method getCanvas {} { return $CCanvasWid }

    method clickCon {aCursor} 
    method clickConClick {a b} 
    method clickDraw {aCursor} 
    method clickDrawClick {a b} 
    method defaultBindings {} 

    method touch {}       ;# sign the map as 'changed'
    method untouch {}     ;# sign the map as 'hasn't changed'
    method hasChanged {}  ;# Q: if the map has changed

    method add {aRoom}     ;# add a room to the Map
    method remRoom {aRoom} ;# add a room to the Map
    method rem {aXY}      ;# rem the room with the coresponding coordinates
    method exists {aXY}   ;# Q: if the room exists

    method getXYs {}      ;# R: return the coordinates of all rooms
    method getRoom {aXY}  ;# RO: return the room with the corresponding coordinate

    method = {source}     ;# copys the contents of another map into this one
    method == {source}    ;# test for equality
    method != {source}    ;# test for equality

    method show {}
    method hide {}
    method createCanvas {}
    private method removeCanvas {}

    method remMapM {aMapM} 

    proc returnMapNameLevelFromFile {aFile}

    method namel {}       ;# returns a string that combines level and name
    method name {} { return $CName }
    method level {} { return $CLevel }
    method id {}    { return $CId }

    method createXYCursor {} {
	$CCanvasWid delete xycursor	
        $CCanvasWid create text 0 0 \
	    -tag xycursor	
    }
    method removeXYCursor {} {
	$CCanvasWid delete xycursor	
    }
    method updateXYCursor {a b} {
	set x    [$CCBWid canvasx $a]
	set y    [$CCBWid canvasy $b]
	set xval [p2mi $x  1]
	set yval [p2mi $y -1]
    
	$CCanvasWid coords        xycursor [expr $x+16] [expr $y-8]
	$CCanvasWid itemconfigure xycursor -text "$xval,$yval"
    }

    proc getNamel   {aName aLevel}
    proc namelname  {aNamel}
    proc namellevel {aNamel}

    proc m2p    {x d} { return [expr ($d*$x*$CGrid)+$CGrid/2] }
    proc m2pmin {x d} { return [expr ($d*($x-0.5)*$CGrid)+$CGrid/2] }
    proc m2pmax {x d} { return [expr ($d*($x+0.5)*$CGrid)+$CGrid/2] }
    proc p2m    {x d} { return [expr $d*($x-($CGrid/2))/$CGrid] }
    proc p2mi   {x d} { return [expr round($d*($x-($CGrid/2))/$CGrid)] }

    method xminpix {} { return [m2pmin $CXmin  1]}
    method xmaxpix {} { return [m2pmax $CXmax  1]}
    method yminpix {} { return [m2pmax $CYmax -1]}
    method ymaxpix {} { return [m2pmin $CYmin -1]}

    method canvasWidth  {} { return [expr [xmaxpix] - [xminpix]] }
    method canvasHeight {} { return [expr [ymaxpix] - [yminpix]] }

    method newZoomPos {a b} 
    method newPlayerPos {a b} 

    method mapSelect {act a b} 
    method unselAll {} 


    method drawGrid {}
    method drawZoomPos {aX aY}
    method drawPlayerPos {aX aY}
    method drawSelection {aX aY} 
    method undrawSelection {aX aY} 
    method remZoomPos {}
    method remPlayerPos {}

    method adjustViewPoint {aX aY}

    proc getMapNLIFromFile {aFile}

    method getConPos {from to} {
	if {[exists [$to getXY]]} {
	    return [[getRoom [$to getXY]] getExitPosTo $from]
	} else {
	    return cc
	}
    }

    method export {aFileName} 
    method save {} 
    method saveMain {fid} 
    method load {} 
    method loadRooms {fid version} 
    method import {aFileName} 
    method loadIfNeeded {}
    method unload {}
}

####################################################################################
class TMapManager {
    # contains the Room interfac
    #          the map selector uif
    #          and the map canvas frame (where the map is drawn)
    # of course also some structs for managing the maps
    inherit itk::Widget


    itk_option define -foreground foreground Foreground black

    variable CParent

    variable CMudderUif
    variable CRoomChangedConfirmer ;# contains a widgets, that will be launched at request

    method getEntryX {} { return [$CMudderUif getEntryX] }

    variable CExitSelector
    variable CClickConMap
    variable CClickMode
    method clickCon {aExitSelector}
    method clickConClick {x y}
    method clickDraw {aExitSelector}
    method clickDrawClick {x y}

    variable CBGLight "\#ffffff"
    variable CBGDef   "\#daceaf"
    variable CBGDark  "\#a97"
    private variable CBGColor "\#daceaf"

    private variable CMud      ;# must have this in order to know where to save all the maps. the actual mud the maps belong to 
       
    private variable CPlayerPos   ;# object pointer TPosition
    private variable CZoomPos     ;# object pointer TPosition

    private variable CMapFrame  ;# widgetID in which the maps should be drawn
    method getMapFrame {} { return $CMapFrame }

    private variable CBGIM
    private variable CFilM
    constructor {aBGIM aMudM aFilM aParent args} {}
    destructor  {}

    private method existsRoomAtPlayer {} {
	return [existsRoom \
		    [$CPlayerPos getMapName]  \
		    [$CPlayerPos getMapLevel] \
		    [$CPlayerPos getX]        \
		    [$CPlayerPos getY]         ]
    }
    private method getRoomAtPlayer {} {
	return [getRoom \
		    [$CPlayerPos getMapName]  \
		    [$CPlayerPos getMapLevel] \
		    [$CPlayerPos getX]        \
		    [$CPlayerPos getY]         ]
    }

    method speedy {aMapName aMapLevel aX aY} {
	# used by TEnryX
	if {[roomHasChanged]} {
	    if {[devMode]} {
		acceptRoom
	    } else {
		$CSes writeMessage "Current room in zoom interface hasnt been saved!"
		return
	    }
	}
	
	if {[existsMap $aMapName $aMapLevel]} {	    
	    speedWalkTo [getMap $aMapName $aMapLevel] $aX $aY
	} else {
	    $CSes writeMessage "There is no such map ('$aMapName', '$aMapLevel')!"
	}
	
    }

    method speedWalkTo {aMap aX aY} {
	if {![existsRoomAtPlayer]} {
	    bell
	    return
	}
	if {![existsRoom [$aMap name] [$aMap level] $aX $aY]} {
	    bell
	    after 300
	    bell
	    after 300
	    bell
	    return
	}

	$CParent setStatusStr "Speedwalk: Search"
	update idletasks
	#puts ===========================================================


	# init
	pqueue yellowpq
	set  start            [getRoomAtPlayer]
	set  dist($start)     0
	set  rmap($start)     [$CPlayerPos getMap]
	set  father($start)   ""
	set  fatherEx($start) ""  
	set  yellow($start)   [yellowpq insert $start $dist($start)]

	set found 0
	# start search
	while {[yellowpq isntEmpty]} {
	    # get the nearest not visited room
	    set nearestYellow         [yellowpq deleteMin]
	    set green($nearestYellow) ""
	    #$CParent setStatusStr "Speedwalk: Search [$rmap($nearestYellow) namel] [$nearestYellow TCoordinate::showval]; distance $dist($nearestYellow)"
	    
	    # test, if to far
	    if {$dist($nearestYellow)>1000} {
		# not found!
		set found 0
		break
	    }
	    # test, if found
	    if { [$aMap == $rmap($nearestYellow)]} {
		if { [$nearestYellow TCoordinate::getX] == $aX } {
		    if { [$nearestYellow TCoordinate::getY] == $aY } {
			#found!
			set found 1
			break
		    }
		}
	    }

	    # markvisited

	    # follow its connections
	    foreach expos [$nearestYellow TExitList::getPoss] {		
		#nodebug puts "  search: $expos"
		set ex [$nearestYellow TExitList::getExit $expos]
		set mn [$ex getMap]
		set ml [$ex getMapLevel]
		set cc [$ex getConnectTo]
		set mx [$cc getX]
		set my [$cc getY]
		
		if {![string compare "" $ml]} {
		    set ml 0
		}
		if {![string compare "" $mn]} {
		    set mn [$rmap($nearestYellow) name]
		    set ml [$rmap($nearestYellow) level]
		}

		#nodebug puts "  $expos: '$mn' -- '$ml'"
		
		# this is special to smm graphs -> connection may lead to void
		if {![existsRoom $mn $ml $mx $my]} { continue }
		set room        [getRoom $mn $ml $mx $my]
		#nodebug puts "      room yes"
		
		# skip if its green
		if {[info exists green($room)]} { continue }

		#nodebug puts "     found yellow: $mn $ml $mx $my"
		if {[info exists yellow($room)]} {
		    # readjust its position, if needed (when dist now is smaller)
		    set nd [expr $dist($nearestYellow) + 1]
		    if { $nd < $dist($room)} {
			# its needed
			set dist($room)     $nd
			set father($room)   $nearestYellow
			set fatherEx($room) $expos
			yellowpq rekey $yellow($room) $dist($room)
			
			set s $yellow($room)
			while {[yellowpq exchangeNeeded $s]} {
			    set f      [yellowpq exchange $s]
			    set yellow([yellowpq getValue $s])   $s
			}
			set yellow($room) $f
				
		    }
		} else {
		    # put it into the pqueue, if its a new yellow
		    # and also remeber its pos in the pqueue
		    set  dist($room)     [expr $dist($nearestYellow) + 1]
		    set  father($room)   $nearestYellow
		    set  fatherEx($room) $expos
		    set  yellow($room)   [yellowpq insert $room $dist($room)]
		    set  rmap($room)     [getMap $mn $ml]
		}
	    }

	}


	if {$found} {
	    $CParent setStatusStr "Speedwalk: Destination found! Building movement list."
	    update idletasks
	    set exl [list]
	    set ar  $nearestYellow
	    while {[string compare "" $fatherEx($ar)]} {
		set exl [linsert $exl 0 $fatherEx($ar)]
		set ar $father($ar)
	    }
	    $CParent setStatusStr "Speedwalking,..."
	    foreach expos $exl {
		useExit $expos
		update idletasks
	    }

	}

	$CParent setStatusStr ""
	::delete object yellowpq
    }

    method getLayout {} {
	set retval [list]
	lappend retval [$itk_component(pw) getFractions]
	lappend retval [$itk_component(zoomUif) getLayout]
	return $retval
    }
    method setLayout {aLayout} {
	set frac [lindex $aLayout 0]
	eval $itk_component(pw) fraction $frac
	$itk_component(zoomUif) setLayout [lindex $aLayout 1]
    }


    method useExit {aPos} 

    method getZif {} { return $itk_component(zoomUif) }

    # just as using the user interface
    method acceptRoom {} {$itk_component(zoomUif) accept} ;# just as pressing the button
    method resetRoom {} {$itk_component(zoomUif) resetit}  ;# just as pressing the button
    method zifRemoveRoom {} { 
	#$itk_component(zoomUif) removeTheRoom
	$itk_component(zoomUif) clearUif
    }
    method zifSetBg {aType aCat aName} {	
	$itk_component(zoomUif) zifSetBg [$CBGIM getObject $aType $aCat $aName]
    }
    method zifSetStuff {aPos aType aCat aName} {	
	$itk_component(zoomUif) zifSetStuff $aPos [$CBGIM getObject $aType $aCat $aName]
    }
    method zifRemStuff {aPos} {
	$itk_component(zoomUif) zifRemStuff $aPos 
    }
    method zifSetText {aText} {	
	$itk_component(zoomUif) zifSetText $aText
    }
    method zifRemText {} {
	$itk_component(zoomUif) zifRemText 
    }
    method zifRemExit {aPos} {
	$itk_component(zoomUif) zifRemExit $aPos
    }
    method zifSetExit {aPos aCommand 
		       aMapName aMapLevel 
		       aConX aConY aDrawX aDrawY
		       aDoor aDoorComOpen aDoorComClose
		       aThickness aColor 
		   } {
	set cc [TCoordinate ::\#auto $aConX $aConY]
	set cd [TCoordinate ::\#auto $aDrawX $aDrawY]

	set ex [TExit ::\#auto \
		    $aPos $aMapName $aMapLevel \
		    $cc $cd \
		    $aCommand \
		    $aDoor $aDoorComOpen $aDoorComClose \
		    $aThickness $aColor ]

	$itk_component(zoomUif) zifSetExit $ex
	::delete object $ex $cc $cd

    }
    

    method devMode {} { return [$CParent getDevMode]}
    method setDevMode {aMode} { return [$CParent setDevMode $aMode]}

    method getBGIM {} { return $CBGIM }

    variable CChangedMapNames
    method startMapping {aMud} ;# "connects" to a Mud -> sets CMud variable
    method startMappingBase {}
    method stopMapping {}
    method mappingStarted {} 
    method reallyStopMapping {aId} 
    method cleanup {}

    private variable CActMap    ;# the actual shown map: object pointer
    method newActMap {aMap}
    method getActMap {}         ;# RO: returns the actual map
    method selActMap {}         ;# get selection of actmap and make it the new act map
    method hasMap {}            ;# 0/1, if a map exist

    method getActCoord {} { return $CZoomPos}        ;# RO: actual coordinate and map -> position
    method getZoomPos {} { return $CZoomPos }        ;	# used by TEntryX
    method getPlayerPos {} { return $CPlayerPos }    ;	# used by TEntryX

    method showMap {} 
    method hideMap {}

    method importMap {aFileName} 
    method reallyImportMap { aFileName aName aLevel }
    method exportMap {aFileName} 
    method saveMap {} 
    method remMap {} 
    method removeMap {} 

    method getMap {aName aLevel} { 
	return [$CMud getMap $aName $aLevel] 
    }
    method existsMap {aName aLevel} { 
	# used by TEntryX
	return [$CMud existsMap $aName $aLevel] 
    }
    method existsRoom {aName aLevel aX aY} {
	# used by TEntryX
	if {[existsMap $aName $aLevel]} {
	    set map [$CMud getMap $aName $aLevel]
	    $map loadIfNeeded
	    set xy  [TCoordinate::GETXY $aX $aY]
	    if {[$map exists $xy]} {
		return 1
	    }
	}
	return 0
    }
    variable CSes
    method assignSession {aSes} { set CSes $aSes }

    method movePlayer {aHow aWhere} {
	if {[roomHasChanged]} {
	    if {[devMode]} {
		acceptRoom
	    } else {
		bell
		$CSes writeMessage "Current room in zoom interface hasnt been saved!"
		return
	    }
	}

	set map [$CPlayerPos getMap]
	set x   [$CPlayerPos getX]
	set y   [$CPlayerPos getY]
	set xy  [TCoordinate::GETXY $x $y]

	#firstof set the zoompos to the playerpos
	if {[$CZoomPos != $CPlayerPos]} {
	    newZoomPos $map $x $y
	} 
	if {![$map exists $xy]} {
	    bell
	    $CSes writeMessage "There is no map!"
	    return
	}
	set room [$map getRoom $xy]	

	switch -- $aHow {
	    useexitwithcommand {
		set aPos [$room getPosOfExitWithCommand $aWhere]
		set sendCommands 1
	    }
	    followexitwithcommand {
		set aPos [$room getPosOfExitWithCommand $aWhere]
		set sendCommands 0
	    }
	    useexit {
		set aPos $aWhere
		set sendCommands 1
	    }	    
	    followexit {
		set aPos $aWhere
		set sendCommands 0
	    }	    
	    default {
		bell
		$CSes writeMessage "No such command!"
		return
	    }
	}



	if {![$room TExitList::exists $aPos]} {
	    bell
	    $CSes writeMessage "No such exit!"
	    return
	}
	
	# get the map to where to go
	set ex [$room getExit $aPos]
	set ml [$ex getMapLevel]
	set mn [$ex getMap]
	if {![string compare "" $ml]} {
	    set ml 0
	}
	if {![string compare "" $mn]} {
	    set mn [$map name]
	    set ml [$map level]
	}

	# walk on the map
	if {[$CMud existsMap $mn $ml]} {
	    # send the command to the mud
	    if {$sendCommands} {
		if {[$ex hasDoor]} {
		    $CSes toSMM [$ex getDoorOpen]
		    $CSes toSMM [$ex getCommand]
		    $CSes toSMM [$ex getDoorClose]
		} else {
		    $CSes toSMM [$ex getCommand]
		}
	    }
	    
	    set nmap [$CMud getMap $mn $ml]
	    set nx   [[$ex getConnectTo] getX]
	    set ny   [[$ex getConnectTo] getY]
	    newZoomPos   $nmap $nx $ny
	    newPlayerPos $nmap $nx $ny
	} else {
	    $CSes writeMessage "The map the exit leads to does not exist! You have to create it before!"
	}
    }

    method getExitWithCommand {aName aLevel aX aY aCom} {
	# used by TEntryX
	if {[existsRoom $aName $aLevel $aX $aY]} {
	    set room [getRoom $aName $aLevel $aX $aY]
	    return [$room getPosOfExitWithCommand $aCom]
	}
	return ""
    }
    method getRoom {aName aLevel aX aY} {
	set map [$CMud getMap $aName $aLevel]
	set xy  [TCoordinate::GETXY $aX $aY]
	return [$map getRoom $xy]
    }
    method setZoomPos {aMapName aMapLevel aX aY} {
	# used by TEntryX
	set map [$CMud getMap $aMapName $aMapLevel]
	newZoomPos $map $aX $aY
    }
    method setPlayerPos {aMapName aMapLevel aX aY} {
	# used by TEntryX
	set map [$CMud getMap $aMapName $aMapLevel]
	newPlayerPos $map $aX $aY
    }
    method changeMap {aMapName aMapLevel} {
	# used by TEntryX
	if {[existsMap $aMapName $aMapLevel]} {
	    newActMap [$CMud getMap $aMapName $aMapLevel]
	}
    }
    method createMap {aMapName aMapLevel} {
	# used by TEntryX
	if {![regexp -- {^[-+]?[0-9]+$} $aMapLevel]} {
	    $CSes writeMessage "Level argument is not integer!"
	    return
	}	
	if {[existsMap $aMapName $aMapLevel]} {
	    $CSes writeMessage "Map already exists!"
	    return
	}	
	reallyCreateNewMap "\{$aMapName\}" "\{$aMapLevel\}"
    }

    method newZoomPos {aMap aX aY} 
    method newPlayerPos {aMap aX aY} 

    private {
	variable _selMap
	variable _selRooms ;# array
	variable _selsx
	variable _selex
	variable _selsy
	variable _seley
	variable _selzx ; #sel zoom pos
	variable _selzy
	variable _selClip
	method   _checkMapSelect {aMap}
	method   _selAdjustLoop {aX aY} {
	    set zoomX [$CZoomPos getX]
	    set zoomY [$CZoomPos getY]
	    if {$zoomX<$aX} {
		set _selsx $zoomX
		set _selex $aX
	    } else {
		set _selsx $aX
		set _selex $zoomX
	    }
	    if {$zoomY<$aY} {
		set _selsy $zoomY
		set _seley $aY
	    } else {
		set _selsy $aY
		set _seley $zoomY
	    }
	    incr _selex
	    incr _seley
	}
    }
    method mapSelect {aMap aAct aX aY} 
    method selClear {} {
	if {[info exist _selMap]} {
	    catch "$_selMap unselAll"
	}
	catch {unset    _selMap}
	catch {unset    _selRooms}
	catch {unset    _selClip}
    }
    method selDelete {} {
	if {![info exist _selMap]} {
	    return
	}
	set h [TRoom ::\#auto]
	foreach xy [array names _selRooms] {
	    if {[$_selMap exists $xy]} {
		regexp {(.*),(.*)} $xy dummy x y
		$h setXY $x $y
		$_selMap remRoom $h
	    }
	}
	::delete object $h
	selClear
    }
    method selFillZif {} {
	if {![info exist _selMap]} {
	    return
	}
	foreach xy [array names _selRooms] {
	}
	selClear
    }
    method selCopy {aClear} {
	if {![info exist _selMap]} {
	    return
	}
	set _selzx [$CZoomPos getX]
	set _selzy [$CZoomPos getY]
	catch {unset _selClip}
	foreach xy [array names _selRooms] {
	    if {[$_selMap exists $xy]} {
		set _selClip($xy) [TRoom ::\#auto [$_selMap getRoom $xy]]
	    } else {
		set _selClip($xy) {}
	    }
	}
	if {$aClear} { selClear }
    }
    method selCut {} {
	selCopy 0
	selDelete
    }
    method selPaste {} {
	set dx  [expr [$CZoomPos getX] - $_selzx]
	set dy  [expr [$CZoomPos getY] - $_selzy]
	set map [$CZoomPos getMap]

	foreach xy [array names _selClip] {
	    if {[string compare {} $_selClip($xy)]} {
		# -- a room to be added
		#    get room
		set h [TRoom ::\#auto $_selClip($xy)]
		#    move to right pos
		$h moveby $dx $dy
		#    adjust exits
		foreach exp [$h TExitList::getPoss] {
		    set ex  [$h TExitList::getExit $exp]
		    if {[$ex isRelative]} {
			set con [$ex getConnectTo]			
			if {[info exists _selClip([$con getXY])]} {
			    $ex moveby $dx $dy
			}
		    }
		}
		#    add to map
		$map add $h
		::delete object $h
	    } else {
		set h [TRoom ::\#auto]
		regexp {(.*),(.*)} $xy dummy x y
		incr x $dx
		incr y $dy
		set XY $x,$y
		if {[$map exists $XY]} {
		    $h setXY $x $y
		    $map remRoom $h
		}
		::delete object $h
	    }
	}
    }

    





    method zoomIsOnActMap {} {
	if {![$CZoomPos exists]} { return 0 }
	if {![hasMap]} { return 0 }
	return [$CActMap == [$CZoomPos getMap]]
    }
    method playerIsOnActMap {} {
	if {![$CPlayerPos exists]} { return 0}
	if {![hasMap]} { return 0}
	return [$CActMap == [$CPlayerPos getMap]]
    }


    method newMapReq {} 
    method resizeMapReq {} 
    method reallyResizeMap {a b c d} 
    method addRoom {aRoom aMap} 
    method remRoom {aRoom aMap} 
    method reallyCreateNewMapAndWalk {aName aLevel aPos}
    method reallyCreateNewMap {aName aLevel} 



    method mapChanged {aMap} {
	set namel [$aMap namel]
	set ls [$itk_component(mapname) size]
	for {set i 0} { $i<$ls } {incr i} {
	    set nnnn [$itk_component(mapname) get $i]
	    set patt "\[*.\] $namel"
	    if {[string match $patt $nnnn]} {
		break
	    }
	}
	sortMapSelector
    }
    method mapUnchanged {aMap} {
	set namel [$aMap namel]
	set ls [$itk_component(mapname) size]
	for {set i 0} { $i<$ls } {incr i} {
	    set nnnn [$itk_component(mapname) get $i]
	    set patt "\[*.\] $namel"
	    if {[string match $patt $nnnn]} {
		break
	    }
	}
	sortMapSelector
    }
    method updateMapsList {} {
	# this is called from outside
	set CActMap ""
	startMappingBase
    }
    method sortMapSelector {} {
	$itk_component(mapname) delete list 0 end
	foreach a [$CMud getAllMapNamesWithIndicator] {
	    $itk_component(mapname) insert list end $a
	}
    }

    method roomHasChanged {} {
	return [$itk_component(zoomUif) hasChanged]
    }
    method informRoomHasChanged {} { $CRoomChangedConfirmer activate }

}

##################################################################################################
##################################################################################################
##################################################################################################
##################################################################################################
##################################################################################################
class TUnsavedMapsUif {
    inherit iwidgets::Shell
    variable CCom
    
    constructor {aCom aNum args} {
	configure -title "Notice"
	configure -modality application
	set CCom $aCom

	set wid [$this childsite]
	
	#########################################################################
	# create message area
 	itk_component add   mframe {frame $wid.mframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(mframe) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0

 	itk_component add   bitm  {label $itk_component(mframe).l -bitmap info } {keep -background}
 	pack $itk_component(bitm) -side left -fill both -expand true -ipadx 20 -ipady 20 -padx 0 -pady 0
 	itk_component add   mess  {
	    message $itk_component(mframe).m -font big -aspect 2000 \
		-text "There is/are $aNum unsaved maps!\nWhat do you want to do?"

	} {keep -background}
 	pack $itk_component(mess) -side left -fill both -expand true -ipadx 20 -ipady 20 -padx 0 -pady 0


	#########################################################################
	# create button
 	itk_component add   butframe {frame $wid.butframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(butframe) -side bottom -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0
 	itk_component add   saveall {
 	    button $itk_component(butframe).saveall -width 26 \
		-text "save all" -font def -command "[code $this closeit s]"
 	} {keep -background}
 	itk_component add   discall {
 	    button $itk_component(butframe).discall -width 26 \
		-text "discard all changes" -font def -command "[code $this closeit d]"
 	} {keep -background}
 	itk_component add   askeach {
 	    button $itk_component(butframe).askeach -width 26 \
		-text "ask for each map separatly" -font def -command "[code $this closeit a]"
 	} {keep -background}
 	pack $itk_component(saveall) $itk_component(discall) $itk_component(askeach) \
	    -fill both -expand true -ipadx 10 -ipady 10 -padx 0 -pady 0 -side left

	eval itk_initialize $args
	activate
    }
    private method closeit {a} {
	eval $CCom $a
	::delete object $this
    }

}
class TSaveDiscardUif {
    inherit iwidgets::Shell
    variable CMapM
    variable CMud
    variable CList
    variable CName
    variable CLevel
    
    constructor {aMapM aMud aList args} {
	configure -title "Notice"
	configure -modality application
	set CMapM $aMapM
	set CMud  $aMud
	set CList $aList

	set wid [$this childsite]
	
	#########################################################################
	# create message area
 	itk_component add   mframe {frame $wid.mframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(mframe) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0

 	itk_component add   bitm  {label $itk_component(mframe).l -bitmap info } {keep -background}
 	pack $itk_component(bitm) -side left -fill both -expand true -ipadx 20 -ipady 20 -padx 0 -pady 0
 	itk_component add   mess  {
	    message $itk_component(mframe).m -font big -aspect 2000 
	} {keep -background}
 	pack $itk_component(mess) -side left -fill both -expand true -ipadx 20 -ipady 20 -padx 0 -pady 0


	#########################################################################
	# create button
 	itk_component add   butframe {frame $wid.butframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(butframe) -side bottom -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0
 	itk_component add   save {
 	    button $itk_component(butframe).saveall -width 16 \
		-text "save changed" -font def -command "[code $this save]"
 	} {keep -background}
 	itk_component add   disc {
 	    button $itk_component(butframe).discall -width 16 \
		-text "discard changes" -font def -command "[code $this discard]"
 	} {keep -background}
 	pack $itk_component(save) $itk_component(disc) \
	    -fill both -expand true -ipadx 10 -ipady 10 -padx 0 -pady 0 -side left

	eval itk_initialize $args
	next
	activate
    }
    private method closeit {} {
	$CMud  remMapM $CMapM
	$CMapM cleanup
	::delete object $this
    }
    method next {} {
	if {[llength $CList]==0} {
	    closeit
	    return
	}
	set nl [lindex $CList 0]
	set CList [lreplace $CList 0 0]
	set CName  [lindex $nl 0]
	set CLevel [lindex $nl 1]
	$itk_component(mess) configure -text "Changes made to map:\nName: $CName\nLevel: $CLevel"
    }
    method save {} {
	[$CMud getMap $CName $CLevel] save
	next
    }
    method discard {} {
	next
    }

}


################################################################
body TMapManager::_checkMapSelect {aMap} {
    if {[info exist _selMap]} {
	if {[$_selMap != $aMap]} {		    
	    unset    _selRooms
	    $_selMap unselAll
	}
    }
}
body TMapManager::mapSelect {aMap aAct aX aY} {
    switch $aAct {
	s {
	    _checkMapSelect        $aMap
	    set _selMap            $aMap
	    if {![info exists _selRooms($aX,$aY)]} {
		set _selRooms($aX,$aY) {}
		$_selMap drawSelection $aX $aY
	    }
	}
	u {
	    _checkMapSelect        $aMap
	    set _selMap            $aMap
	    if {[info exists _selRooms($aX,$aY)]} {
		unset _selRooms($aX,$aY) 
		$_selMap undrawSelection $aX $aY
	    }
	}
	t {
	    _checkMapSelect        $aMap	    
	    set _selMap            $aMap
	    if {[info exists _selRooms($aX,$aY)]} {
		unset _selRooms($aX,$aY) 
		$_selMap undrawSelection $aX $aY
	    } else {
		set _selRooms($aX,$aY) {}
		$_selMap drawSelection $aX $aY
	    }
	}
	sr {
	    _checkMapSelect        $aMap
	    set _selMap            $aMap
	    _selAdjustLoop         $aX $aY
	    for {set x $_selsx} {$x<$_selex} {incr x} {
		for {set y $_selsy} {$y<$_seley} {incr y} {
		    if {![info exists _selRooms($x,$y)]} {
			set _selRooms($x,$y) {}
			$_selMap drawSelection $x $y
		    }
		}
	    }	    			     
	}
	ur {
	    _checkMapSelect        $aMap
	    set _selMap            $aMap
	    _selAdjustLoop         $aX $aY
	    for {set x $_selsx} {$x<$_selex} {incr x} {
		for {set y $_selsy} {$y<$_seley} {incr y} {
		    if {[info exists _selRooms($x,$y)]} {
			unset _selRooms($x,$y) 
			$_selMap undrawSelection $x $y
		    }
		}
	    }	    			     
	}
	tr {
	    _checkMapSelect        $aMap
	    set _selMap            $aMap
	    _selAdjustLoop         $aX $aY
	    for {set x $_selsx} {$x<$_selex} {incr x} {
		for {set y $_selsy} {$y<$_seley} {incr y} {
		    if {$x==$aX && $y==$aY} { continue }
		    if {[info exists _selRooms($x,$y)]} {
			unset _selRooms($x,$y) 
			$_selMap undrawSelection $x $y
		    } else {
			set _selRooms($x,$y) {}
			$_selMap drawSelection $x $y
		    }
		}
	    }
	    
	}
    }
}
body TMapManager::newZoomPos {aMap aX aY} {
    if {[roomHasChanged]} {
	if {[devMode]} {
	    acceptRoom
	} else {
	    informRoomHasChanged
	    return
	}
    }

    $CZoomPos setXY $aX $aY
    if { [$aMap != [$CZoomPos getMap]] } {
	[$CZoomPos getMap] remZoomPos
	$CZoomPos setMap $aMap
	newActMap $aMap
    }

    if {[$aMap exists [$CZoomPos getXY]]} {
	$itk_component(zoomUif) setUif   [$aMap getRoom [$CZoomPos getXY]] [getActMap]
    } else {
	$itk_component(zoomUif) resetUif $aMap $aX $aY
    }
    $aMap drawZoomPos $aX $aY
}
body TMapManager::newPlayerPos {aMap aX aY} {
    if {[roomHasChanged]} {
	if {[devMode]} {
	    acceptRoom
	} else {
	    informRoomHasChanged
	    return
	}
    }

    $CPlayerPos setXY $aX $aY
    if { [$aMap != [$CPlayerPos getMap]] } {
	[$CPlayerPos getMap] remPlayerPos
	$CPlayerPos setMap $aMap
    }
    $aMap drawPlayerPos $aX $aY
}
body TMapManager::newMapReq {} {
    if {![mappingStarted]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information \
	    "No Session has been started!
           \nYou need to have a active session in order to use mapping."
	return
    }
    $itk_component(newmap) configure -command "[code $this reallyCreateNewMap \{%n\} \{%l\}]"
    $itk_component(newmap) launchUif
#     TNewMapUif $itk_interior.newmap \
# 	-command "[code $this reallyCreateNewMap \{%n\} \{%l\}]" \
# 	-master [winfo toplevel $itk_interior] \
# 	-tops [winfo toplevel $itk_interior] 
}
body TMapManager::resizeMapReq {} {
    if {![hasMap]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information \
	    "Somehow there is no map,..\ndoesn't it sound logic to have a map before trying to resize it?!?"
	return
    }
    TResizeMapUif $itk_interior.resizemap -size [$CActMap size] \
	-command "[code $this reallyResizeMap %xmin %xmax %ymin %ymax]" \
	-master [winfo toplevel $itk_interior] \
	-tops [winfo toplevel $itk_interior] 
}
body TMapManager::reallyResizeMap {a b c d} {
    $CActMap resize $a $b $c $d
    return 1
}
body TMapManager::addRoom {aRoom aMap} {
    if {![mappingStarted]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information \
	    "Funny,.. and where is your map?!\nStart a session first and create/activate a map."
	return

    }
    if {[string compare "" $aMap]==0} { 
	$CActMap add $aRoom
    } else {
	$aMap add $aRoom
    }
}
body TMapManager::remRoom {aRoom aMap} {
    if {![mappingStarted]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information \
	    "Funny,.. and where is your map?!\nStart a session first and create/activate a map."
	return

    }
    if {[string compare "" $aMap]==0} { 
	$CActMap remRoom $aRoom 
    } else {
	$aMap remRoom $aRoom 
    }
}
##############################3
body TMapManager::reallyCreateNewMapAndWalk {aName aLevel aPos} {
    reallyCreateNewMap $aName $aLevel
    useExit $aPos
}
body TMapManager::reallyCreateNewMap {aName aLevel} {
    regexp "^{(.*)}$" $aName  h name
    regexp "^{(.*)}$" $aLevel h level

    if {![$CMud addMap $name $level]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "The map '$name' (level: $level)\n already exists!"
	return 0
    }

    set map [$CMud getMap $name $level]    
    $map touch

    sortMapSelector

    newActMap $map
    
    if {[expr ![$CZoomPos exists]]} {
	$CZoomPos setMap      $map
	$CZoomPos setXY       0 0
	$map      drawZoomPos 0 0
    }
    if {[expr ![$CPlayerPos exists]]} {
	$CPlayerPos setMap $map
	$CPlayerPos setXY 0 0
	update
	$map        drawPlayerPos 0 0
    }
    return 1
}
body TMapManager::useExit {aPos} {    
    if {[roomHasChanged]} {
	if {[devMode]} {
	    acceptRoom
	} else {
	    informRoomHasChanged
	    return
	}
    }

    set map [$CPlayerPos getMap]
    set x   [$CPlayerPos getX]
    set y   [$CPlayerPos getY]
    set xy  [TCoordinate::GETXY $x $y]

    #firstof set the zoompos to the playerpos
    if {[$CZoomPos != $CPlayerPos]} {
	newZoomPos $map $x $y
    } 
    if {![$map exists $xy]} {
	bell
	return
    }
    set room [$map getRoom $xy]	
    if {![$room TExitList::exists $aPos]} {
	bell
	return
    }
    
    # get the map to where to go
    set ex [$room getExit $aPos]
    set ml [$ex getMapLevel]
    set mn [$ex getMap]
    if {![string compare "" $ml]} {
	set ml 0
    }
    if {![string compare "" $mn]} {
	set mn [$map name]
	set ml [$map level]
    }

    # walk on the map
    if {[$CMud existsMap $mn $ml]} {
	# send the command to the mud
	if {[$ex hasDoor]} {
	    $CSes toSMM [$ex getDoorOpen]
	    $CSes toSMM [$ex getCommand]
	    $CSes toSMM [$ex getDoorClose]
	} else {
	    $CSes toSMM [$ex getCommand]
	}
	
	set nmap [$CMud getMap $mn $ml]
	set nx   [[$ex getConnectTo] getX]
	set ny   [[$ex getConnectTo] getY]
	newZoomPos   $nmap $nx $ny
	newPlayerPos $nmap $nx $ny
    } else {
	TConfirmer $itk_interior.useexit -foreground $itk_option(-foreground)\
	    -command "[code $this reallyCreateNewMapAndWalk \{$mn\} \{$ml\}] $aPos" \
	    -question "Do you want to create the following map:\nname: $mn\nlevel: $ml?" 
    }
}

body TMapManager::importMap {aFileName} {
    set nl [TMap::returnMapNameLevelFromFile $aFileName]
    
    $itk_component(newmap) configure -command "[code $this reallyImportMap $aFileName \{%n\} \{%l\}]"
    $itk_component(newmap) configure -name [lindex $nl 0] 
    $itk_component(newmap) configure -level [lindex $nl 1]
    $itk_component(newmap) launchUif

#     TNewMapUif $itk_interior.newmap \
# 	-name [lindex $nl 0] -level [lindex $nl 1] \
# 	-command "[code $this reallyImportMap $aFileName \{%n\} \{%l\}]" \
# 	-master [winfo toplevel $itk_interior] \
# 	-tops [winfo toplevel $itk_interior] 
}

body TMapManager::reallyImportMap { aFileName aName aLevel } {
    regexp "^{(.*)}$" $aName  h name
    regexp "^{(.*)}$" $aLevel h level

    if {![$CMud importMap $aFileName $name $level]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "The map '$name' (level: $level)\n already exists!"
	return 0
    }

    set map [$CMud getMap $name $level]    
    $map touch

    sortMapSelector

    newActMap $map
    
    if {[expr ![$CZoomPos exists]]} {
	$CZoomPos setMap      $map
	$CZoomPos setXY       0 0
	$map      drawZoomPos 0 0
    }
    if {[expr ![$CPlayerPos exists]]} {
	$CPlayerPos setMap $map
	$CPlayerPos setXY 0 0
	update
	$map        drawPlayerPos 0 0
    }
    return 1
}
body TMapManager::exportMap {aFileName} {
    if {[string match "" $CActMap]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "No map available!"
	return
    }

    $CActMap export $aFileName
}
body TMapManager::saveMap {} {
    if {[string match "" $CActMap]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "No map available!"
	return
    }

    $CActMap save  
}
body TMapManager::removeMap {} {
    if {[zoomIsOnActMap]} { $CZoomPos setNull } 
    if {[playerIsOnActMap]} { $CPlayerPos setNull } 
    $CMud deleteMap $CActMap
}
body TMapManager::remMap {} {
    if {[string match "" $CActMap]} {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "No map available!"
	return
    }
    TConfirmer $itk_interior.remmap -command "[code $this removeMap]" \
	-question "Really remove actual map?"
}
body TMapManager::reallyStopMapping {aId} {
    switch $aId {
	s {
	    foreach ml $CChangedMapNames {
		[$CMud getMap [lindex $ml 0] [lindex $ml 1]] save
	    }
	    $CMud remMapM $this
	    cleanup
	}
	d {
	    $CMud remMapM $this
	    cleanup
	}
	a {
	    TSaveDiscardUif $itk_interior.savedis $this $CMud $CChangedMapNames \
		-master [winfo toplevel $itk_interior] \
		-tops [winfo toplevel $itk_interior] 
	}
    }
}
body TMapManager::cleanup {} {
    global TV
    set CMud    ""
    set CMapDir ""
    set CActMap ""
    $itk_component(mapname) delete list  0 end
    set TV($this,mapname) ""
    $CZoomPos setNull
    $CPlayerPos setNull
    $itk_component(zoomUif) resetUif "" 0 0    
    catch {unset    _selMap}
    catch {unset    _selRooms}
}
body TMapManager::stopMapping {} {
    if {[mappingStarted]} {
	set CChangedMapNames [$CMud getChangedMapNames]
	set l [llength $CChangedMapNames]

	if {$l>0} {
	    TUnsavedMapsUif $itk_interior.unsavedmaps [code $this reallyStopMapping] $l \
		-master [winfo toplevel $itk_interior] \
		-tops [winfo toplevel $itk_interior] 		
	    return
	} 
	$CMud remMapM $this
    } 
    cleanup
}
body TMapManager::mappingStarted {} {
    return [expr ![string match "" $CMud]]
}
body TMapManager::hasMap {} {
    return [expr ![string match "" $CActMap]]
}
body TMapManager::startMappingBase {} {
    set maps [$CMud getAllMaps]	
    
    if {[llength $maps] >0} { 
	sortMapSelector
	
	set map [lindex $maps 0]	
	newActMap $map	
	if {[expr ![$CZoomPos exists]]} {
	    $CZoomPos setMap      $map
	    $CZoomPos setXY       0 0
	    $map      drawZoomPos 0 0
	}
	if {[expr ![$CPlayerPos exists]]} {
	    $CPlayerPos setMap $map
	    $CPlayerPos setXY 0 0
	    update
	    $map        drawPlayerPos 0 0
	}
	newPlayerPos $map 0 0
	newZoomPos   $map 0 0
    } else {
	$itk_component(mapname) configure -editable 1
	$itk_component(mapname) clear
	$itk_component(mapname) clear entry
	set TV($this,mapname) ""
	$itk_component(mapname) configure -editable 0
    }
}
body TMapManager::startMapping {aMud} {
    if {![TMapManager::mappingStarted]} {
	set CMud    $aMud
	$CMud addMapM $this

	startMappingBase
    } else {
	TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
	    -information "There is already a mud active!"
    }
}
configbody TMapManager::foreground {
    #$itk_component(mapname) configure -foreground $itk_option(-foreground) 
}
body TMapManager::constructor {aBGIM aMudderUif aFilM aParent args} {
    set CExitSelector ""
    set CClickConMap  ""
    set CClickMode    "normal"

    set CParent     $aParent
    set CMud        ""
    set CMapDir     ""
    set CActMap     ""
    set CZoomPos    [TPosition ::\#auto]
    set CPlayerPos  [TPosition ::\#auto]
    set CBGIM       $aBGIM
    set CMudderUif  $aMudderUif
    set CFilM       $aFilM

    # newmap gui
    itk_component add  newmap {
    TNewMapUif $itk_interior.newmap \
	-master [winfo toplevel $itk_interior] \
	-tops [winfo toplevel $itk_interior] 
    } {
	keep -background
	keep -foreground
    }    


    # mainframe
    #itk_component add mainframe {
    #frame $itk_interior.mainf -background $CBGColor -bd 2 -relief ridge
    #} {keep -background}
    #pack  $itk_component(mainframe) -side top -expand 1 -fill both -ipadx 0 -ipady 0 -padx 0 -pady 0

    itk_component add   pw {
	iwidgets::panedwindow $itk_interior.mainf -background $CBGColor \
	    -orient horizontal
    } {keep -background}
    pack $itk_component(pw) -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0 -side left

    $itk_component(pw) add  zif    -margin 0
    $itk_component(pw) add  maps   -margin 0


    # zoom IF/RoomUif
    itk_component add zoomUif {
	TRoomUif [$itk_component(pw) childsite zif].zoom $this $CBGIM -background $CBGColor 
    } {
	keep -background
	keep -foreground
    }
    pack  $itk_component(zoomUif) -side top -expand 1 -fill both -ipadx 0 -ipady 0 -padx 0 -pady 0

    itk_component add secframe {
	frame [$itk_component(pw) childsite maps].secframe -background $CBGColor -bd 0 -relief ridge
    } {
	keep -background
    }
    pack  $itk_component(secframe) -side top -expand 1 -fill both -ipadx 0 -ipady 0 -padx 0 -pady 0


    # mapname
    itk_component add mapnamef {
	frame $itk_component(secframe).mapname -background $CBGColor -bd 2 -relief ridge
    } {keep -background}
    pack  $itk_component(mapnamef) -side top -expand 0 -fill x -ipadx 0 -ipady 0 -padx 0 -pady 0
    itk_component add   mapname {
	iwidgets::combobox $itk_component(mapnamef).cb \
	    -popupcursor hand1 -listheight 100  \
	    -textbackground $CBGLight -selectbackground $CBGDark \
	    -labeltext "Map: " -textfont def -labelfont defBold \
	    -selectioncommand [code $this selActMap] -editable 0 \
	    -textvariable TV($this,mapname) 
    } {
	keep -background
	keep -foreground
    }
    pack  $itk_component(mapname) -side top -expand 0 -fill x -ipadx 0 -ipady 0 -padx 0 -pady 0
 
    # mapframe
    itk_component add mapframe {
	frame $itk_component(secframe).map -background $CBGColor -bd 2 -relief ridge
    } {
	keep -background
    }
    pack  $itk_component(mapframe) -side top -expand 1 -fill both -ipadx 0 -ipady 0 -padx 0 -pady 0
    set CMapFrame  $itk_component(mapframe)

    #----- no showuppers
    #- Room Changed confirmer
    set CRoomChangedConfirmer [TInformer $itk_interior.rcc -foreground black \
				   -hide hide -vanishafter 1 -autoactivate 0 \
				   -information "Room changed! Either 'accept' or 'reset' first."]
    addFGCom "$CRoomChangedConfirmer configure -foreground %c" 
    eval itk_initialize $args

    $itk_component(pw) fraction 14 86

}
body TMapManager::destructor {} {
}

body TMapManager::clickCon {aExitSelector} {
    set CClickMode con
    set CExitSelector $aExitSelector
    set CClickConMap  [getActMap]
    focus $itk_component(secframe)
    grab $itk_component(secframe)
    [getActMap] clickCon "hand2 black"
}
body TMapManager::clickConClick {x y} {
    set CClickMode normal
    set am [getActMap]
    grab  release $itk_component(secframe)
    if {[$CClickConMap == $am]} {
	$CExitSelector clickConClick $x $y "" ""
    } else {
	newActMap $CClickConMap
	$CExitSelector clickConClick $x $y [$am name] [$am level]
    }
    set CExitSelector ""
}
body TMapManager::clickDraw {aExitSelector} {
    set CClickMode draw
    set CExitSelector $aExitSelector
    set CClickConMap  [getActMap]
    focus $itk_component(secframe)
    grab $itk_component(secframe)
    [getActMap] clickDraw "hand2 black"
}
body TMapManager::clickDrawClick {x y} {
    set CClickMode normal
    set am [getActMap]
    grab  release $itk_component(secframe)
    if {[$CClickConMap == $am]} {
	$CExitSelector clickDrawClick $x $y 
    } else {
	newActMap $CClickConMap
	$CExitSelector clickDrawClick $x $y 
    }
    set CExitSelector ""
}
body TMapManager::selActMap {} {
    global TV
    
    set nnnn  [string range $TV($this,mapname) 2 end]
    if {![string compare "" $nnnn]} {
	return
    }

    set name  [TMap::namelname  $nnnn]
    set level [TMap::namellevel $nnnn]
    set map   [$CMud getMap $name $level]
    
    switch -- $CClickMode {
	con {
	    [getActMap] defaultBindings
	    newActMap $map
	    [getActMap] clickCon "hand2 black"
	    [getActMap] adjustViewPoint [$CZoomPos getX] [$CZoomPos getY]
	}
	draw {
	    [getActMap] defaultBindings
	    newActMap $map
	    [getActMap] clickDraw "hand2 black"
	    [getActMap] adjustViewPoint [$CZoomPos getX] [$CZoomPos getY]
	}
	default {
	    newActMap $map
	}
    }
}

body TMapManager::newActMap {aMap} {
    global TV

    if {![string match "" $CActMap]} {
	TMapManager::hideMap
    }
    
    set CActMap $aMap
    $CActMap loadIfNeeded

    set TV($this,mapname) [$aMap namel]

    TMapManager::showMap  
}





body TMapManager::getActMap {} {
    return $CActMap
}

body TMapManager::showMap {} {
    $CActMap show
}
body TMapManager::hideMap {} {
    $CActMap hide
}


##########################################################

body TMap::export {aFileName} {
    set       fid [open $aFileName "w+"]
    saveMain $fid
    close    $fid
}
body TMap::save {} {
    set       fid [open [$CMud getMapFile $CId] "w+"]
    saveMain $fid
    close    $fid
    untouch
}
body TMap::saveMain {fid} {
    # do saving here
    puts $fid "2" ; # version
    puts $fid $CId
    puts $fid $CName
    puts $fid $CLevel
    puts $fid $CXmin
    puts $fid $CXmax
    puts $fid $CYmin
    puts $fid $CYmax

    foreach rpos [getXYs] {
	[getRoom $rpos] save $fid
    }
}


body TMap::getMapNLIFromFile {aFile} {
    set fid [open $aFile "r"]
    gets $fid version
    gets $fid id
    gets $fid name
    gets $fid level
    close $fid    
    return [list $name $level $id]
}
body TMap::loadIfNeeded {} {
    if {$CUnLoaded} {
	load
	set CUnLoaded 0
    }
}
body TMap::returnMapNameLevelFromFile {aFile} {
    ## map version changes affect here too!!!!!
    set fid [open $aFile "r"]
    # 
    gets $fid version
    gets $fid id
    gets $fid name
    gets $fid level

    close $fid

    return [list $name $level]
}
body TMap::load {} {
    set fid [open [$CMud getMapFile $CId] "r"]
    # do loading here
    gets $fid version
    gets $fid CId
    gets $fid CName
    gets $fid CLevel
    gets $fid CXmin
    gets $fid CXmax
    gets $fid CYmin
    gets $fid CYmax

    loadRooms $fid $version
    
    close $fid
}
body TMap::import {aFileName} {
    set fid [open $aFileName "r"]
    # do loading here
    gets $fid version
    gets $fid fileId
    gets $fid fileName
    gets $fid fileLevel
    gets $fid CXmin
    gets $fid CXmax
    gets $fid CYmin
    gets $fid CYmax

    loadRooms $fid $version
    
    close $fid
}
body TMap::loadRooms {fid version} {
    switch -- $version {
	1 {
	    createCanvas	    
	    while {[gets $fid line]>0} {
		set room [TRoom ::\#auto]
		$room load_V1ConvertToV2 $line [[$CMud getMapM] getBGIM]
		add $room
		::delete object $room
	    } 
	    touch
	}
	2 {
	    createCanvas	    
	    while {[gets $fid line]>0} {
		set room [TRoom ::\#auto]
		$room load $line [[$CMud getMapM] getBGIM]
		add $room
		::delete object $room
	    } 
	}
	default {
	    error "unknown version for map: [namel]"
	}
    }
}



body TMap::namelname {aNamel} {
    if {[regexp -- {(.+); Level: ([\+|\-])(\ *)([0-9]+)$} "$aNamel" all name sign sp val]} {
	return $name
    } else {
	return $aNamel
    }
}
body TMap::namellevel {aNamel} {
    if {[regexp -- {(.+); Level: ([\+|\-])(\ *)([0-9]+)$} "$aNamel" all name sign sp val]} {
	if {[string match "+" $sign]} {
	    return [expr 1*$val]
	} else {
	    return [expr -1*$val]
	}
    } else {
	return 0
    }
}
body TMap::namel {} {
    if {$CLevel==0} {
	return $CName
    } else {
	if {$CLevel>0} {
	    set level [format "%3d" $CLevel]
	    return "$CName; Level: +$level"
	} else {
	    set level [format "%3d" [expr -1*$CLevel]]
	    return "$CName; Level: -$level"
	}
    }
}
body TMap::getNamel {aName aLevel} {
    if {$aLevel==0} {
	return $aName
    } else {
	if {$aLevel>0} {
	    set level [format "%3d" $aLevel]
	    return "$aName; Level: +$level"
	} else {
	    set level [format "%3d" [expr -1*$aLevel]]
	    return "$aName; Level: -$level"
	}
    }
}
body TMap::drawZoomPos {aX aY} {
    remZoomPos
    
    $CCanvasWid create rectangle \
	[expr [m2pmin $aX 1]-0] [expr [m2pmin $aY -1]+0] \
	[expr [m2pmax $aX 1]+0] [expr [m2pmax $aY -1]-0] \
	-outline red -width 3 -tag zoompos
    $CCanvasWid lower zoompos grid
    set CHasZoom 1
    set CZoomX   $aX
    set CZoomY   $aY
}

body TMap::drawSelection {aX aY} {
    $CCanvasWid create rectangle \
	[expr [m2pmin $aX 1]+5] [expr [m2pmin $aY -1]-5] \
	[expr [m2pmax $aX 1]-5] [expr [m2pmax $aY -1]+5] \
	-outline yellow -width 2 -tag "sel sel$aX,$aY"
    #$CCanvasWid lower "sel$aX,$aY" grid
}
body TMap::undrawSelection {aX aY} {
    $CCanvasWid delete "sel$aX,$aY"
}
body TMap::unselAll {} {
    $CCanvasWid delete sel
}

body TMap::drawPlayerPos {aX aY} {
    remPlayerPos
    
    $CCanvasWid create rectangle \
	[expr [m2pmin $aX 1]+1] [expr [m2pmin $aY -1]-1] \
	[expr [m2pmax $aX 1]-1] [expr [m2pmax $aY -1]+1] \
	-outline blue -width 3 -tag playerpos
    catch "$CCanvasWid lower playerpos zoompos"

    adjustViewPoint $aX $aY

    set CHasPlayer 1
    set CPlayerX   $aX
    set CPlayerY   $aY
}
body TMap::adjustViewPoint {aX aY} {
    set xp [winfo width  $CCBWid]
    set yp [winfo height $CCBWid]
    set Xp [expr double($CXmax - $CXmin)*$CGrid]
    set Yp [expr double($CYmax - $CYmin)*$CGrid]
    set xa [expr double($aX - $CXmin)*$CGrid]
    set ya [expr double($CYmax - $aY)*$CGrid]
    set xf [expr ($xa - $xp/2)/$Xp]
    set yf [expr ($ya - $yp/2)/$Yp]

    $CCBWid xview moveto $xf
    $CCBWid yview moveto $yf
}
body TMap::remZoomPos {} {
    $CCanvasWid delete zoompos
    set CHasZoom 0    
}
body TMap::remPlayerPos {} {
    ::catch "$CCanvasWid delete playerpos"
    set CHasPlayer 0
}
body TMap::drawGrid {} {
    
    $CCanvasWid delete grid
    
    #draw horizontal lines in x direction
    set start [expr $CYmin]
    set ende  [expr $CYmax+2]
    for {set i $start} {$i<$ende} {incr i 1} {
        $CCanvasWid create line [xminpix] [m2pmin $i -1] [xmaxpix] [m2pmin $i -1] \
	    -fill grey70 -tag grid
    }
    #draw vertical lines in y direction
    set start [expr $CXmin]
    set ende  [expr $CXmax +2]
    for {set i $start} {$i < $ende} {incr i 1} {
        $CCanvasWid create line [m2pmin $i 1] [yminpix] [m2pmin $i 1] [ymaxpix] \
	    -fill grey70 -tag grid
    }
    $CCanvasWid lower grid
}

body TMap::mapSelect {act a b} {
    set x [p2mi [$CCBWid canvasx $a]  1]
    set y [p2mi [$CCBWid canvasy $b] -1]
    $CMapM mapSelect $this $act $x $y
}
body TMap::newZoomPos {a b} {
    if {[$CMapM roomHasChanged]} {
	if {[$CMapM devMode]} {
	    $CMapM acceptRoom
	} else {
	    $CMapM informRoomHasChanged
	    return
	}
    }

    set x [p2mi [$CCBWid canvasx $a]  1]
    set y [p2mi [$CCBWid canvasy $b] -1]
    $CMapM newZoomPos $this $x $y
    drawZoomPos $x $y

}
body TMap::newPlayerPos {a b} {
    if {[$CMapM roomHasChanged]} {
	if {[$CMapM devMode]} {
	    $CMapM acceptRoom
	} else {
	    $CMapM informRoomHasChanged
	    return
	}
    }

    set x [p2mi [$CCBWid canvasx $a]  1]
    set y [p2mi [$CCBWid canvasy $b] -1]
    $CMapM newPlayerPos $this $x $y
    drawPlayerPos $x $y

}
body TMap::remMapM {aMapM} {
    set CMapM ""
    TMap::removeCanvas
    TMap::unload
}
body TMap::resize {a b c d} {
    set CXmin $a
    set CXmax $b
    set CYmin $c
    set CYmax $d

    TMap::save
    TMap::removeCanvas
    TMap::unload

    TMap::loadIfNeeded
    $CMapM newActMap $this

    if {$CHasZoom} {
	drawZoomPos $CZoomX $CZoomY
    }
    if {$CHasPlayer} {
	drawPlayerPos $CPlayerX $CPlayerY
    }

}
body TMap::unload {} {
    if {!$CUnLoaded} {
	foreach xy [TMap::getXYs] {
	    TMap::rem $xy
	}
	set CUnLoaded 1
	untouch
    }
}
body TMap::removeCanvas {} {
    if {[info exists CCanvasWid]} {
	destroy $CCanvasWid
	unset CCanvasWid
    }
}
body TMap::createCanvas {} {
    if {[info exists CCanvasWid]} {
	return
    }
    set CMapM      [$CMud getMapM]
    set CBGIM      [$CMapM getBGIM]
    set CParentWid [$CMapM getMapFrame]
    set CCanvasWid "$CParentWid.mapc$CId"

    iwidgets::scrolledcanvas $CCanvasWid \
	-relief ridge -sbwidth 10 \
	-background $CBGColor -textbackground $CBGColor -selectbackground $CBGColor
    set CCBWid [$CCanvasWid childsite]
    
    TMap::drawGrid
    
    # cursor
    bind    $CCBWid <Enter>  [code $this createXYCursor]
    bind    $CCBWid <Leave>  [code $this removeXYCursor]
    bind    $CCBWid <Motion> [code $this updateXYCursor %x %y]
    # selections 
    bind    $CCBWid <<MapSelect>>               "[code $this mapSelect s  %x %y]"
    bind    $CCBWid <<MapSelectRegion>>         "[code $this mapSelect sr %x %y]"
    bind    $CCBWid <<MapUnSelect>>             "[code $this mapSelect u  %x %y]"
    bind    $CCBWid <<MapUnSelectRegion>>       "[code $this mapSelect ur %x %y]"
    bind    $CCBWid <<MapToggleSelect>>         "[code $this mapSelect t  %x %y]"
    bind    $CCBWid <<MapToggleSelectRegion>>   "[code $this mapSelect tr %x %y]"
    # primary actions
    bind    $CCBWid <<MapPrimary>>     "[code $this newZoomPos %x %y]"
    bind    $CCBWid <<MapPlacePlayer>> "[code $this newPlayerPos %x %y]"
    bind    $CCBWid <<MapPowerwalk>>   "[code $this speedWalkTo %x %y]"
}
body TMap::clickCon {aCursor} {
    $CCBWid config -cursor $aCursor
    bind    $CCBWid <<MapPrimary>> "[code $this clickConClick %x %y]"
}
body TMap::clickConClick {a b} {
    set x [p2mi [$CCBWid canvasx $a]  1]
    set y [p2mi [$CCBWid canvasy $b] -1]
    $CMapM clickConClick $x $y
    defaultBindings
}
body TMap::clickDraw {aCursor} {
    $CCBWid config -cursor $aCursor
    bind    $CCBWid <<MapPrimary>> "[code $this clickDrawClick %x %y]"
}
body TMap::clickDrawClick {a b} {
    set x [p2mi [$CCBWid canvasx $a]  1]
    set y [p2mi [$CCBWid canvasy $b] -1]
    $CMapM clickDrawClick $x $y
    defaultBindings
}
body TMap::defaultBindings {} {
    $CCBWid config -cursor {}
    bind    $CCBWid <<MapPrimary>> "[code $this newZoomPos %x %y]"
}
body TMap::constructor {args} {
    switch [llength $args] {
	5 {
	    # this is called for new created maps?
	    set CZoomX     ""
	    set CZoomY     ""
	    set CPlayerX   ""
	    set CPlayerY   ""
	    set CHasZoom   0
	    set CHasPlayer 0

	    set CName      [lindex $args 0]
	    set CLevel     [lindex $args 1]
	    set CId        [lindex $args 2]
	    set CMud       [lindex $args 3]
	    set CUnLoaded  [lindex $args 4]
	    
	    set CMapM [list]
	    untouch
	}
	7 {
	    set sw [lindex $args 0]
	    switch -- $sw {
		"import" {
		    set CZoomX     ""
		    set CZoomY     ""
		    set CPlayerX   ""
		    set CPlayerY   ""
		    set CHasZoom   0
		    set CHasPlayer 0

		    set fn         [lindex $args 1]
		    set CName      [lindex $args 2]
		    set CLevel     [lindex $args 3]
		    set CId        [lindex $args 4]
		    set CMud       [lindex $args 5]	    
		    set CMapM      [lindex $args 6]

		    set CUnLoaded  0
		    import $fn
		    touch
		}
		default {
		    error "Wrong numer of arguments to constructor! BP2"
		}
	    }
	}
	default {
	    error "Wrong numer of arguments to constructor!"
	}
	
    }	
}

body TMap::destructor {} {
    foreach pos [TMap::getXYs] {
	::delete object [TMap::getRoom $pos]
    }	
    removeCanvas
}
body TMap::touch {} {
    set CChanged 1
    foreach mm $CMapM {
	$CMapM mapChanged $this
    }
}
body TMap::untouch {} {
    set CChanged 0
    foreach mm $CMapM {
	$CMapM mapUnchanged $this
    }
}
body TMap::hasChanged {} {return $CChanged}
body TMap::add {aRoom} {
    # will get their touch with calling 'rem'
    set xy [$aRoom getXY]
    TMap::rem $xy
    set CRoomList($xy) [TRoom ::\#auto $aRoom]
    $aRoom draw $CCanvasWid $this
}
body TMap::remRoom {aRoom} {
    # will get their touch with calling 'rem'
    set xy [$aRoom getXY]
    set x  [$aRoom getX]
    set y  [$aRoom getY]
    TMap::rem $xy
    set CRoomList($xy) [TRoom ::\#auto]
    $CRoomList($xy) setXY $x $y
    
    $CRoomList($xy) drawRem $CCanvasWid $this
    TMap::rem $xy
}
body TMap::rem {aXY} {
    if {[TMap::exists $aXY]} {
	::delete object $CRoomList($aXY)
	unset CRoomList($aXY)
    }
    if {!$CUnLoaded} {
	# only mark as touched, if map is not first newly loaded!
	touch
    }
}
body TMap::exists {aXY} {
    return [info exists CRoomList($aXY)]
}
body TMap::getXYs {} {
    return [array names CRoomList]
}
body TMap::getRoom {aXY} {
    return $CRoomList($aXY)
}
body TMap::== {source} {
    if {[string match $CName [$source name]] && [string match $CLevel [$source level]]} {
	return 1
    } else {
	return 0
    }
}
body TMap::!= {source} {
    if {[string match $CName [$source name]] && [string match $CLevel [$source level]]} {
	return 0
    } else {
	return 1
    }
}
body TMap::= {source} {
    foreach xy [$source getXYs] {
	TMap::add [$source getRoom $xy]
    }
}

body TMap::show {} {
    createCanvas
    pack $CCanvasWid -fill both -expand true
}
body TMap::hide {} {
    pack forget $CCanvasWid
}


#################################################################################
class TNewMapUif {
    inherit iwidgets::Texteddialog

    itk_option define -command command Command "return 1" {}
    itk_option define -name    name    Name    "main" {}
    itk_option define -level   level   Level   "0" {}
    # the command given to the command option _must_ return a boolean!
    #  -> 0: uif stays
    #  -> 1: = success, uif goes

    constructor {args} {
	configure -foreground black
	addFGCom [code $this configure -foreground %c] 

	global TV
	set TV($this,level) 0
	set TV($this,name) ""

	configure -title "Create New Map"
	configure -modality application
	configure -text "Please, enter a name and a level for the map!"
	configure -aspect 333
	buttonconfigure OK     -text "Create" -command "[code $this oki]"
	buttonconfigure Cancel -text "Cancel" -command "[code $this canc]"

	#addFGCom [code $this buttonconfigure OK       -foreground %c] 
	#addFGCom [code $this buttonconfigure Cancel   -foreground %c] 


	#########################################################################
	itk_component add   name {
	    iwidgets::entryfield [childsite].name -labeltext "Name: " \
		-width 20 -labelmargin 0 -textvariable TV($this,name)
	} {
	    keep -background
	    keep -foreground
	}
	itk_component add   level {
	    iwidgets::entryfield [childsite].level -labeltext "Level: " -validate integer \
		-width 4 -labelmargin 0 -textvariable TV($this,level)
	} {
	    keep -background
	    keep -foreground
	}
	pack $itk_component(name) $itk_component(level) -side top -fill both -expand 1
	iwidgets::Labeledwidget::alignlabels $itk_component(name) $itk_component(level) 

	eval itk_initialize $args
	#launchUif
    } 
    private method canc {}        { 
	global TV
	unset TV($this,name)
	unset TV($this,level)
	deactivate
	#::delete object $this
    }
    private method oki {}        { 
	global TV

	set name [string trim $TV($this,name)]
	if {![string compare "" $name]} {
	    TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
		-information "You have to give a 'name' for the map!"	    
	    return
	} 
	scan $TV($this,level) "%d" level
	if {![info exists level]} {
	    TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
		-information "You have to give a 'level' for the map!"	    
	    return
	} 
	
	regsub -all {%n} $itk_option(-command) $name helper
	regsub -all {%l} $helper $level theCom
	if {[eval $theCom]} {
	    canc
	}
    }
    method launchUif {}   { clearUif ; activate }
    private method clearUif  {}   { 
	global TV
	set TV($this,name)  $itk_option(-name)
	set TV($this,level) $itk_option(-level) 
    }

}



#################################################################################
class TResizeMapUif {
    inherit iwidgets::Texteddialog

    itk_option define -command command Command "return 1" {}
    itk_option define -size    size    Size    "0 0 0 0"  {}
    # the command given to the command option _must_ return a boolean!
    #  -> 0: uif stays
    #  -> 1: = success, uif goes

    constructor { args } {} {
	global _TResizeMapUif
	configure -title "Resize Map"
	configure -modality application
	configure -text "Please, enter the new dimensions of the map.\nResizing will also save the map automatically!"
	configure -aspect 333
	buttonconfigure OK     -text "Resize" -command "[code $this oki]"
	buttonconfigure Cancel -text "Cancel" -command "[code $this canc]"

	#########################################################################
	foreach aaa {xmin xmax ymin ymax} {
	    itk_component   add $aaa { 
		iwidgets::spinint  [childsite].$aaa \
		    -labeltext "[string range $aaa 0 0] [string range $aaa 1 end]: " \
		    -step 5 -width 5 -validate integer \
		    -increment [code $this increm $aaa] -decrement [code $this decrem $aaa] \
		    -arroworient horizontal -textvariable _TResizeMapUif($this,$aaa)
	    } {keep -background}
	}

	pack $itk_component(xmin) $itk_component(xmax) $itk_component(ymin) $itk_component(ymax) \
	    -side top -fill both -expand 1
	iwidgets::Labeledwidget::alignlabels \
	    $itk_component(xmin) $itk_component(xmax) $itk_component(ymin) $itk_component(ymax) 


	eval itk_initialize $args

	set _TResizeMapUif($this,xmin) [lindex $itk_option(-size) 0]
	set _TResizeMapUif($this,xmax) [lindex $itk_option(-size) 1]
	set _TResizeMapUif($this,ymin) [lindex $itk_option(-size) 2]
	set _TResizeMapUif($this,ymax) [lindex $itk_option(-size) 3]

	activate 
    } 
    method increm {aaa} {
	global _TResizeMapUif
	if {![string compare "" $_TResizeMapUif($this,$aaa)]} {
	    set _TResizeMapUif($this,$aaa) 0
	}
	incr _TResizeMapUif($this,$aaa) 5
    }
    method decrem {aaa} {
	global _TResizeMapUif
	if {![string compare "" $_TResizeMapUif($this,$aaa)]} {
	    set _TResizeMapUif($this,$aaa) 0
	}
	incr _TResizeMapUif($this,$aaa) -5
    }
    private method canc {}        { 
	global _TResizeMapUif
	unset _TResizeMapUif($this,xmin)
	unset _TResizeMapUif($this,xmax)
	unset _TResizeMapUif($this,ymin)
	unset _TResizeMapUif($this,ymax)
	::delete object $this
    }
    private method oki {}        { 
	global _TResizeMapUif
	set theCom $itk_option(-command)
	foreach aaa {xmin xmax ymin ymax} {
	    if {![string compare "" $_TResizeMapUif($this,$aaa)]} {
		set _TResizeMapUif($this,$aaa) 0
	    }
	}
	
	regsub -all {%xmin} $theCom $_TResizeMapUif($this,xmin) theCom
	regsub -all {%xmax} $theCom $_TResizeMapUif($this,xmax) theCom
	regsub -all {%ymin} $theCom $_TResizeMapUif($this,ymin) theCom
	regsub -all {%ymax} $theCom $_TResizeMapUif($this,ymax) theCom
	if {[eval $theCom]} {
	    canc
	}
    }


}









