class TMudManager {
    private common CMudList  ;# is an array of objects
    private common CFilM     ;# the file system manager

    constructor {aFileSystemManager args} {
	set CFilM $aFileSystemManager

	foreach mf [$CFilM getAllMudDataFiles] {
	    set mud [TMud ::\#auto $CFilM $mf]
	    set CMudList([$mud name]) $mud
	}
    }
    method add {aName aHost aPort} {
	if {[exists $aName]} {
	    ::delete object $CMudList($aName)
	    unset CMudList($aName)	
	}
	set id [getNewId]
	set CMudList($aName) [TMud ::\#auto $CFilM $aName $aHost $aPort $id]
    }
    method rem {aName} {
	if {[exists $aName]} {
	    $CMudList($aName) wipeout
	    ::delete object $CMudList($aName)
	    unset CMudList($aName)	
	}
    }
    method getNewId {} {
	set id 0
	set idlist [getMudIds]
	while {[lsearch -exact $idlist $id]!= -1} {
	    incr id
	}
	return $id
    }
    method getMudIds {} { 
	set idlist [list]
	foreach name [getMudNames] {
	    lappend idlist [[getMud $name] id]
	}
	return $idlist
    }
    method getMudNames {} { 
	return [lsort [array names CMudList]] 
    }
    method getMud {aName} { return $CMudList($aName) } 
    method exists {aName} { return [info exists CMudList($aName)] } 
}

class TMudManagerUif {
    inherit iwidgets::Shell

    variable CMudM
    variable CWid
    variable CWinWid

    constructor {aMudM args} {
	global TV
	set CMudM $aMudM

	configure -title "Mud Manager"
	configure -modality application

	set CWid [$this childsite]
	regsub -- {^::} $this {} CWinWid
	
	#########################################################################
	# create buttons
 	itk_component add   butframe {frame $CWid.butframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(butframe) -side bottom -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0

 	itk_component add   ok {
 	    button $itk_component(butframe).ok \
		-width 10 -text "save" -command "[code $this save]" -fg black
 	} {keep -background}
 	itk_component add   rem {
 	    button $itk_component(butframe).rem \
		-width 10 -text "remove" -command "[code $this remove]" -fg black
 	} {keep -background}
 	itk_component add   canc {
 	    button $itk_component(butframe).canc \
		-width 10 -text "close" -command "[code $this canc]"  -fg black 
 	} {keep -background}
 	pack $itk_component(ok) $itk_component(rem) $itk_component(canc) \
 	    -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0 -side left
	addFGCom [code $itk_component(ok) configure -fg %c] 
	addFGCom [code $itk_component(rem) configure -fg %c] 
	addFGCom [code $itk_component(canc) configure -fg %c] 


	#########################################################################
	# mud interface
 	itk_component add   mframe {frame $CWid.mframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(mframe) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0

	######### name
	itk_component add   name {
	    iwidgets::entryfield $itk_component(mframe).name -labeltext "name: " \
		-width 20 -labelmargin 0 -textvariable TV($this,name) \
		-foreground black
	} {keep -background}
	pack $itk_component(name) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0
	######### host and port
	itk_component add   hostport {
	    frame $itk_component(mframe).hp -relief flat -bd 0 
	} {keep -background}
	pack $itk_component(hostport) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0
	
	itk_component add   host {
	    iwidgets::entryfield $itk_component(hostport).host -labeltext "host: " \
		-width 20 -labelmargin 0 -textvariable TV($this,host) \
		-foreground black
	} {keep -background}
	itk_component add   port {
	    iwidgets::entryfield $itk_component(hostport).port -labeltext "port: " \
		-width 4 -validate integer -labelmargin 0 -textvariable TV($this,port) \
		-foreground black
	} {keep -background}
	pack $itk_component(host) $itk_component(port) \
	    -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0 -side left

	addFGCom [code $itk_component(name) configure -foreground %c] 
	addFGCom [code $itk_component(host) configure -foreground %c] 
	addFGCom [code $itk_component(port) configure -foreground %c] 

	#########################################################################
	# muds selector
 	itk_component add   msframe {frame $CWid.msframe -relief ridge -bd 2 } {keep -background}
 	pack $itk_component(msframe) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0

	itk_component add   mudselector {
	    iwidgets::scrolledlistbox $itk_component(msframe).ms \
		-labeltext "Muds: \n\n\n\n\n\n\n\n\n\n" -labelpos w \
		-selectioncommand "[code $this selected]" -height 1 \
		-foreground black
	} {keep -background}
 	pack $itk_component(mudselector) -side top -fill both -expand true -ipadx 0 -ipady 0 -padx 0 -pady 0
	addFGCom [code $itk_component(mudselector) configure -foreground %c] 

	##########################################################################
	# align all that stuff
	iwidgets::Labeledwidget::alignlabels \
	    $itk_component(mudselector) $itk_component(host) $itk_component(name) 

	updateUifList
	clearUif
	eval itk_initialize $args	
    }
    method exists      {aName} { return [$CMudM exists $aName] }
    method getMudNames {}      { return [$CMudM getMudNames] }
    method updateUifList {} {
	$itk_component(mudselector) delete 0 end
	foreach mn [getMudNames] {
	    $itk_component(mudselector) insert end $mn
	}
    }
    method save {} {
	global TV
	set name [string trim $TV($this,name)]
	if {![string compare $name ""]} {
	    TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
		-information "You have to give a 'name' for the mud!"
	    return
	}

	if {[exists $name]} {
	    TConfirmer $itk_interior.save -foreground $itk_option(-foreground)\
		-command "[code $this add $name $TV($this,host) $TV($this,port)]" \
		-question "The mud '$name' already exists!\nDo You really want to overide it?" 
	    return
	}
	add $name $TV($this,host) $TV($this,port)
    }
    method remove {} {
	global TV
	set name [string trim $TV($this,name)]
	if {![exists $name]} {
	    TInformer $itk_interior.mess -foreground $itk_option(-foreground)\
		-information "A Mud with the name '$name' does not exist!"
	    return
	}
	TConfirmer $itk_interior.remove -foreground $itk_option(-foreground)\
	    -command "[code $this rem $name]" \
	    -question "Do you really want to remove the mud '$name'?\nYou will loose all information - including maps e.g.." 
	
    }
    method add {aName aHost aPort} {
	$CMudM add $aName $aHost $aPort 
	updateUifList
    }
    method rem {aName} {
	$CMudM rem $aName
	updateUifList
	clearUif
    }
    method selected {} {
	global TV
	set name [$itk_component(mudselector) getcurselection]
	if {![string compare "" $name]} { return }
	set mud  [$CMudM getMud $name]
	set TV($this,name) $name
	set TV($this,host) [$mud host]
	set TV($this,port) [$mud port]
    }
    method clearUif {} {global TV; set TV($this,name) "";set TV($this,host) "";set TV($this,port) "" }
    method launchUif {}   { clearUif ; activate }
    method canc {}        { deactivate }
}

#################################################################
class TMud {
    variable CFilM
    variable CName    ;# must be unique
    variable CHost
    variable CPort
    # CId: must be unique too: this is needed to request files and directories for the mud data to be stored
    #  => may never change again! 
    variable CId      
    variable CMapM

    variable CMapArray  ;# this is an array of objects

    constructor {args} {} {
	set CMapM "" ;# must be set later, when a the mud is assigned to a mapmanager
	switch [llength $args] {
	    2 {
		# this is called, if a mud is loaded at start
		set CFilM [lindex $args 0]
		readFile  [lindex $args 1]
		foreach mf [$CFilM getAllMapFiles $CId] {
		    set nli [TMap::getMapNLIFromFile $mf]
		    addMapMain [lindex $nli 0] [lindex $nli 1] [lindex $nli 2] 1
		}
	    }
	    5 {
		# this is called, if a mud is generated new
		set CFilM [lindex $args 0]
		set CName [lindex $args 1]
		set CHost [lindex $args 2]
		set CPort [lindex $args 3]
		set CId   [lindex $args 4]
		createFileSystem
	    }
	    default {
		error "Wrong number of arguments to constructor!"
	    }
	}
    }
    method addMapM {aMapM} { set CMapM $aMapM}
    method remMapM {aMapM} { 
	foreach map [getAllMaps] {
	    $map remMapM $CMapM
	}
	set CMapM ""	
    }
    method getMapM {}    { return $CMapM } 
    ############################################################
    method show {} { puts "$CId -- $CName || $CHost $CPort" }
    method name {} { return $CName }
    method host {} { return $CHost }
    method port {} { return $CPort }
    method id   {} { return $CId }
    ############################################################
    method getAllMapFiles {} {
	return [$CFilM getAllMapFiles $CId]
    }
    method getMapFile {aId} {
	return [$CFilM getMapFile $CId $aId]
    }
    method removeFileSystem {} { 
	file delete -force -- [$CFilM getMudDir    $CId]
    }
    method createFileSystem {} {
	file mkdir    [$CFilM getMudDir    $CId]
	file mkdir    [$CFilM getMudMapDir $CId]

	set    fid [open [$CFilM getMudDataFile $CId] "w+"]
	puts  $fid 1 ;# file version 
	puts  $fid [id]
	puts  $fid [name]
	puts  $fid [host]
	puts  $fid [port]
	close $fid
    }
    method readFile {aFile} {
	set    fid [open $aFile "r"]
	gets  $fid version
	gets  $fid CId
	gets  $fid CName
	gets  $fid CHost
	gets  $fid CPort
	close $fid
    }
    method wipeout {} {
	removeFileSystem
    }

    #############################################################
    method importMap {aFileName aName aLevel}  {
	# add a _new_ map to the MUD
	set aLevel [expr $aLevel]
	if {[existsMap $aName $aLevel]} {
	    return 0
	} 
	set CMapArray($aLevel|$aName) \
	    [TMap ::\#auto import $aFileName $aName $aLevel [getNewMapId] $this [getMapM]]
	return 1
    }
    method addMap {aName aLevel}  {
	# add a _new_ map to the MUD
	set aLevel [expr $aLevel]
	if {[existsMap $aName $aLevel]} {
	    return 0
	} 
	addMapMain $aName $aLevel [getNewMapId] 0
	return 1
    }
    method addMapMain {aName aLevel aId aMustLoad}  {
	set CMapArray($aLevel|$aName) [TMap ::\#auto $aName $aLevel $aId $this $aMustLoad]
    }
    method deleteMap {aMap} {
	set n [$aMap name] 
	set l [$aMap level]
	set i [$aMap id]
	remMap $n $l
	file delete -force -- [getMapFile $i]
	foreach mm $CMapM {
	    $mm updateMapsList
	}
    }
    method remMap {aName aLevel} {
	# rem the Map
	set aLevel [expr $aLevel]
	if {[existsMap $aName $aLevel]} {
	    ::delete object $CMapArray($aLevel|$aName)
	    unset CMapArray($aLevel|$aName)
	}
    }
    method existsMap {aName aLevel} {
	# Q: if the Map exists
	set aLevel [expr $aLevel]
	return [info exists CMapArray($aLevel|$aName)]
    }
    method getAllMaps {} {
	set ml [list]
	foreach mm [array names CMapArray] {
	    lappend ml $CMapArray($mm)
	}
	return $ml
    }
    method getAllMapNamesWithIndicator {} {
	set namellist   [list]
	foreach mm [array names CMapArray] {
	    lappend namellist [$CMapArray($mm) namel]
	}
	set ll [list]
	foreach namel [lsort $namellist] {
	    if { [[getMapFromNamel $namel] hasChanged] } {
		lappend  ll "* $namel"
	    } else {
		lappend  ll ". $namel"
	    }
	}
	return $ll
    }
    method getMapNames {} {
	# R: return the {name level} list of all maps
	set nl   [array names CMapArray]
	set li   [list]    
	foreach mn $nl {
	    regexp -- {(-[0-9]+|\+[0-9]+|[0-9]+)\|(.+)} $mn dummy a b
	    lappend li [list $b $a]
	}
	return $li
    }
    method getChangedMapNames {} {
	# R: return the {name level} list of all maps
	set nl   [array names CMapArray]
	set li   [list]    
	foreach mn $nl {
	    if {[$CMapArray($mn) hasChanged]} {
		regexp -- {(-[0-9]+|\+[0-9]+|[0-9]+)\|(.+)} $mn dummy a b
		lappend li [list $b $a]
	    }
	}
	return $li
    }
    method getMapFromNamel {aNamel} {
	return [getMap [TMap::namelname $aNamel] [TMap::namellevel $aNamel]]
    }
    method getMap {aName aLevel} {
	set aLevel [expr $aLevel]
	# RO: return the corresponding map
	return $CMapArray($aLevel|$aName)
    }
    
    #-----------
    method getAllMapIds {} { 
	set idlist [list]
	foreach mn [getMapNames] {
	    set name  [lindex $mn 0]
	    set level [lindex $mn 1]
	    lappend idlist [[getMap $name $level] id]
	}
	return $idlist
    }
    method getNewMapId {} {
	set id 0
	set idlist [getAllMapIds]
	while {[lsearch -exact $idlist $id]!= -1} {
	    incr id
	}
	return $id
    }

}


#################################################################








