#----------------------------------------------------
# evaluate command line parameters
for {set i 0} {$i<$argc} {incr i} {
    set clp [lindex $argv $i]
    switch -- $clp {
	"-mapwindow" -
	"-mw" {
	    set SMMMAPWINDOW 1
	}
	"-nomapwindow" -
	"-nmw" {
	    set SMMMAPWINDOW 0
	}
	default {
	    puts "ERROR: '$clp' unknown command line parameter"
	    puts "Must be one of the following:"
	    puts "  -mapwindow, -mw"
	    puts "  -nomapwindow, -nmw"
	    exit
	}
    }
}
#----------------------------------------------------
global SMM
set SMM(version) 5.0
if {[info exists DEVMODE]} {
    set SMM(devmode) 1
} else {
    set SMM(devmode) 0
}

wm title . "SMM++"
message .smmstartupmsg -bg white -aspect 1500 -justify center \
    -text "Starting SMM++ (Version $SMM(version)).\nLoading libraries and precompiling code.\nPlease, be patient."
pack .smmstartupmsg 
update

#----------------------------------------------------
# adjust the user dir, if requested
if {[info exists env(SMMPATH)]} {
    set SMMPATH $env(SMMPATH)
}

#----------------------------------------------------
lappend auto_path [file join $SMMSRCPATH sources]
lappend auto_path [file join $SMMSRCPATH sources mudding]
lappend auto_path [file join $SMMSRCPATH sources mapping]
lappend auto_path [file join $SMMSRCPATH sources rsc]
lappend auto_path [file join $SMMSRCPATH sources itcl iwidgets generic]
lappend auto_path [file join $SMMSRCPATH sources editor]
lappend auto_path [file join $SMMSRCPATH sources error]


if {[info exists TCLPATH]} {
    set SMMIWIDGETSPATH [file join $TCLPATH lib iwidgets3.0.0]
}
if {[info exists SMMIWIDGETSPATH]} {
    lappend auto_path $SMMIWIDGETSPATH
}


package require Iwidgets 3.0
source [file join $SMMSRCPATH sources itcl iwidgets generic line.tcl]

source [file join $SMMSRCPATH sources itcl iwidgets generic shell.tcl]
source [file join $SMMSRCPATH sources itcl iwidgets generic dialogshell.tcl]
source [file join $SMMSRCPATH sources itcl iwidgets generic texteddialog.tcl]
source [file join $SMMSRCPATH sources itcl iwidgets generic labeledwidget.tcl]
source [file join $SMMSRCPATH sources itcl iwidgets generic panedwindow.tcl]

source [file join $SMMSRCPATH sources itcl iwidgets generic fontconfig.tcl]
source [file join $SMMSRCPATH sources itcl iwidgets generic fontmanager.tcl]

source [file join $SMMSRCPATH sources itcl iwidgets generic colormanager.tcl]



if {[info exists TROUBLE(optionmenu)]} {
    source [file join $SMMSRCPATH bin optionmenu.itk]
}

#----------------------------------------------------
#----------------------------------------------------

#----------------------------------------------------
#----------------------------------------------------
#TLoadFeedback .smmLoadmeter 157 $SMMSRCPATH 
#proc loadmeter  {aString} { .smmLoadmeter nextStatus $aString }
#proc loadmeter2 {aString} { .smmLoadmeter nextSubStatus $aString }
proc loadmeter  {aString} {  }
proc loadmeter2 {aString} {  }

#----------------------------------------------------
loadmeter "Building filesystem manager."
TFileSystemManager film $SMMPATH $SMMSRCPATH

#----------------------------------------------------
loadmeter "Building GUI manager."
TGUI  guim film
[guim getColorManager] add "FG def GUI" FGGuiDef black 2
proc addFGCom {com} {
    [guim getColorManager] addCommand FGGuiDef $com
}
guim addGUIColors

#---------------------------------------------------
loadmeter "Building version info."
set    SMM(vinfstr) "Version:     $SMM(version)\n"
append SMM(vinfstr) "User path:   $SMMPATH\n"
append SMM(vinfstr) "SMM sources: $SMMSRCPATH\n"
append SMM(vinfstr) "Tcl version: $tcl_version\n"
append SMM(vinfstr) "Tk version:  $tk_version\n"
append SMM(vinfstr) "\n"
append SMM(vinfstr) "Platform:\n"
append SMM(vinfstr) "Byteorder:   $tcl_platform(byteOrder)\n"
append SMM(vinfstr) "Machine:     $tcl_platform(machine)\n"
append SMM(vinfstr) "OS:          $tcl_platform(os)\n"
append SMM(vinfstr) "OS version:  $tcl_platform(osVersion)\n"
append SMM(vinfstr) "Platform:    $tcl_platform(platform)"

#----------------------------------------------------
loadmeter "Building mud manager."
TMudManager        mudm film

loadmeter "Building session manager."
TSesManager        sesm film mudm

#---------------------------------------------------
# create the image manager
loadmeter "Building image manager."
TImageManager imam [file join $SMMSRCPATH images]
loadmeter "Loading bg images (bw/ontop)."
imam  ddfLoad      bg_bw_ontop
loadmeter "Loading bg images (bw/below)."
imam  ddfLoad      bg_bw_below
loadmeter "Loading bg images (color/ontop)."
imam  ddfLoad      bg_color_ontop
loadmeter "Loading bg images (color/below)."
imam  ddfLoad      bg_color_below
loadmeter "Loading system images."
imam  ddfLoad      system
loadmeter "Loading stuff images 1."
imam  ddfLoad      stuff
loadmeter "Loading stuff images 2."
imam  ddfLoad      stuff_XL

#----------------------------------------------------
# TExitManager: defines some common look up behaviour
#  for all exits -> must be created just once
loadmeter "Building exit manager."
TExitManager ExitManager 32 imam

#----------------------------------------------------
# the error logger
loadmeter "Building Error Logger"
TErrorLogger .err
proc msgError {aStr} { .err msgError $aStr }
#----------------------------------------------------
# create the main UIF
loadmeter "Building main user interface."
TMainUif .smm guim film mudm sesm imam $SMM(vinfstr) $SMMMAPWINDOW .err

# at last load user color definitions
guim loadColorrc

#----------------------------------------------------

#.smmLoadmeter ready
wm withdraw .
destroy .smmstartupmsg


